# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-07-16 13:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("game", "0069_remove_user_levels_from_episodes"),
    ]

    operations = [
        migrations.AlterField(
            model_name="level",
            name="character_name",
            field=models.CharField(
                blank=True,
                choices=[
                    ("Van", "Van"),
                    ("Dee", "Dee"),
                    ("Nigel", "Nigel"),
                    ("Kirsty", "Kirsty"),
                    ("Wes", "Wes"),
                    ("Phil", "Phil"),
                ],
                default=None,
                max_length=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="level",
            name="theme_name",
            field=models.CharField(
                blank=True,
                choices=[
                    ("grass", "grass"),
                    ("snow", "snow"),
                    ("farm", "farm"),
                    ("city", "city"),
                ],
                default=None,
                max_length=10,
                null=True,
            ),
        ),
    ]



# Generated by Django 3.2.25 on 2024-06-24 12:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('game', '0093_alter_level_character_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='level',
            name='hint',
            field=models.TextField(default='Think back to earlier levels. What did you learn?', max_length=10000),
        ),
        migrations.AddField(
            model_name='level',
            name='lesson',
            field=models.TextField(default='Can you find the shortest route?', max_length=10000),
        ),
        migrations.AddField(
            model_name='level',
            name='subtitle',
            field=models.TextField(blank=True, max_length=100, null=True),
        ),
    ]



from game.end_to_end_tests.base_game_test import BaseGameTest

from selenium.webdriver.common.by import By

class TestLanguageDropdown(BaseGameTest):
    def test_level_language_dropdown(self):
        page = self.go_to_level(1)
        assert page.element_exists_by_id("language_dropdown")

        self.selenium.find_element(By.ID, "language_dropdown").click()
        self.selenium.find_element(By.ID, "language_dropdown_fr").click()

        text_count = len(self.selenium.find_elements(By.XPATH, ("//*[contains(text(),'Move forwards')]")))
        assert text_count == 0



from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select, WebDriverWait

from game.end_to_end_tests.base_game_test import BaseGameTest
from game.views.level_editor import available_blocks

DELAY_TIME = 10


class TestLevelEditor(BaseGameTest):
    def set_up_basic_map(self):
        self.test_level_editor_displays()

        add_road_button = self.selenium.find_element(By.ID, "add_road")
        add_road_button.click()

        road_start = self.selenium.find_element(
            By.CSS_SELECTOR, "rect[x='130'][y='530']"
        )
        road_end = self.selenium.find_element(By.CSS_SELECTOR, "rect[x='330'][y='530']")
        ActionChains(self.selenium).drag_and_drop(road_start, road_end).perform()

        mark_start_button = self.selenium.find_element(By.ID, "start")
        mark_start_button.click()
        ActionChains(self.selenium).move_to_element(road_start).click().perform()

        add_house_button = self.selenium.find_element(By.ID, "add_house")
        add_house_button.click()
        ActionChains(self.selenium).move_to_element(road_end).click().perform()

        return [road_start, road_end]

    def test_level_editor_displays(self):
        page = self.go_to_level_editor()

        assert page.element_exists_by_id("paper")

    def test_code_tab_blocks_load(self):
        page = self.go_to_level_editor()
        page.go_to_code_tab()

        block_types = [block.type for block in available_blocks()]

        for block_type in block_types:
            assert page.element_exists_by_id(f"{block_type}_checkbox")
            assert page.element_exists_by_id(f"{block_type}_image")

    def test_multiple_houses(self):
        [road_start, road_end] = self.set_up_basic_map()

        road_middle = self.selenium.find_element(
            By.CSS_SELECTOR, "rect[x='230'][y='530']"
        )

        add_house_button = self.selenium.find_element(By.ID, "add_house")
        add_house_button.click()
        ActionChains(self.selenium).move_to_element(road_middle).click().perform()

        added_houses = self.selenium.find_elements(
            By.CSS_SELECTOR, "rect[fill='#0000ff']"
        )
        assert len(added_houses) == 2

        delete_house_button = self.selenium.find_element(By.ID, "delete_house")
        delete_house_button.click()
        ActionChains(self.selenium).move_to_element(road_middle).click().perform()
        ActionChains(self.selenium).move_to_element(road_start).perform()

        houses_after_delete = self.selenium.find_elements(
            By.CSS_SELECTOR, "rect[fill='#0000ff']"
        )
        assert len(houses_after_delete) == 1

    def test_cow_on_origin(self):
        page = self.go_to_level_editor()
        [road_start, road_end] = self.set_up_basic_map()

        origin_space = self.selenium.find_elements(
            By.CSS_SELECTOR, "rect[fill='#ff0000']"
        )
        assert len(origin_space) == 1

        page.go_to_scenery_tab()

        draggable_cow = self.selenium.find_element(By.ID, "cow")
        ActionChains(self.selenium).click_and_hold(draggable_cow).move_to_element(
            road_start
        ).perform()
        start_space_warning = self.selenium.find_elements(
            By.CSS_SELECTOR,
            "rect[fill='#e35f4d'][fill-opacity='0.7'][x='130'][y='530']",
        )
        assert len(start_space_warning) == 1

    def test_cow_on_house(self):
        page = self.go_to_level_editor()
        [road_start, road_end] = self.set_up_basic_map()

        house_space = self.selenium.find_elements(
            By.CSS_SELECTOR, "rect[fill='#0000ff']"
        )
        assert len(house_space) == 1
        assert road_end == house_space[0]

        page.go_to_scenery_tab()

        draggable_cow = self.selenium.find_elements(By.ID, "cow")
        assert len(draggable_cow) == 1
        ActionChains(self.selenium).click_and_hold(draggable_cow[0]).move_to_element(
            road_end
        ).perform()
        allowed_space = self.selenium.find_elements(
            By.CSS_SELECTOR, "rect[fill='#87e34d']"
        )
        assert len(allowed_space) == 0

    def test_draggable_decor(self):
        page = self.go_to_level_editor()
        page.go_to_scenery_tab()

        source_tree = self.selenium.find_element(By.ID, "tree2")
        end_space = self.selenium.find_element(
            By.CSS_SELECTOR, "rect[x='130'][y='530']"
        )
        ActionChains(self.selenium).drag_and_drop(source_tree, end_space).perform()

        decor_tree = self.selenium.find_elements(By.CSS_SELECTOR, "image[x='0'][y='0']")
        cloned_source_tree = self.selenium.find_elements(By.ID, "tree2")
        assert len(decor_tree) == 1
        assert len(cloned_source_tree) == 1

    def test_draggable_cow(self):
        page = self.go_to_level_editor()
        page.go_to_scenery_tab()

        source_cow = self.selenium.find_element(By.ID, "cow")
        end_space = self.selenium.find_element(
            By.CSS_SELECTOR, "rect[x='130'][y='530']"
        )
        ActionChains(self.selenium).drag_and_drop(source_cow, end_space).perform()

        scenery_cow = self.selenium.find_elements(
            By.CSS_SELECTOR, "image[x='0'][y='0']"
        )
        cloned_source_cow = self.selenium.find_elements(By.ID, "cow")
        assert len(scenery_cow) == 1
        assert len(cloned_source_cow) == 1

    def test_draggable_traffic_light(self):
        page = self.go_to_level_editor()
        page.go_to_scenery_tab()

        source_light = self.selenium.find_element(By.ID, "trafficLightRed")
        end_space = self.selenium.find_element(
            By.CSS_SELECTOR, "rect[x='130'][y='530']"
        )
        ActionChains(self.selenium).drag_and_drop(source_light, end_space).perform()

        scenery_light = self.selenium.find_elements(
            By.CSS_SELECTOR, "image[x='0'][y='0']"
        )
        cloned_source_light = self.selenium.find_elements(By.ID, "trafficLightRed")
        assert len(scenery_light) == 1
        assert len(cloned_source_light) == 1

    def test_custom_description_and_hint(self):
        # login
        self.login_once()

        # go to level editor and set up basic map
        page = self.go_to_level_editor()
        [road_start, road_end] = self.set_up_basic_map()

        # fill in custom description and hint fields
        page.go_to_description_tab()
        self.selenium.find_element(By.ID, "subtitle").send_keys("test subtitle")
        self.selenium.find_element(By.ID, "description").send_keys("test description")

        page.go_to_hint_tab()
        self.selenium.find_element(By.ID, "hint").send_keys("test hint")

        # save level and choose to play it
        page.go_to_save_tab()
        self.selenium.find_element(By.ID, "levelNameInput").send_keys("test level")
        self.selenium.find_element(By.ID, "saveLevel").click()
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.presence_of_element_located((By.ID, "play_button"))
        )
        self.selenium.find_element(By.ID, "play_button").click()

        # check to see if custom subtitle and lesson appear on the initial popup
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.presence_of_element_located((By.ID, "myModal-lead"))
        )
        modal_text = self.selenium.find_element(By.ID, "myModal-lead").get_attribute(
            "innerHTML"
        )
        assert "test subtitle" in modal_text
        assert "test description" in modal_text
        self.selenium.find_element(By.ID, "close-modal").click()

        # wait for modal to disappear
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.none_of(
                EC.visibility_of_all_elements_located((By.ID, "myModal-mainText"))
            )
        )

        # check to see if the custom hint appears on failure modal
        fast_tab = self.selenium.find_element(By.ID, "fast_tab")
        fast_tab.click()
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.visibility_of_element_located((By.ID, "hintPopupBtn"))
        )
        hint_button = self.selenium.find_element(By.ID, "hintPopupBtn")
        hint_button.click()

        hint_modal_text = self.selenium.find_element(By.ID, "hintText").get_attribute(
            "innerHTML"
        )
        hint_modal_style = self.selenium.find_element(By.ID, "hintText").get_attribute(
            "style"
        )
        assert "display: none" in hint_button.get_attribute("style")
        assert "display: block" in hint_modal_style
        assert "test hint" in hint_modal_text

        self.selenium.find_element(By.ID, "try_again_button").click()

        # check to see if the custom hint appears on the hint popup
        self.selenium.find_element(By.ID, "help_tab").click()
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.visibility_of_element_located((By.ID, "myModal-mainText"))
        )
        hint_modal_text_two = self.selenium.find_element(
            By.ID, "myModal-mainText"
        ).get_attribute("innerHTML")
        assert "test hint" in hint_modal_text_two

    def test_solar_panels(self):
        """test that the solar panels appear as a scenery option when clicking on the scenery tab
        and that they disappear as a scenery option when switching to an incompatible theme, i.e. snow
        """
        page = self.go_to_level_editor()
        page.go_to_scenery_tab()

        solar_panel_style = self.selenium.find_element(
            By.ID, "solar_panel"
        ).get_attribute("style")
        assert "inline" in solar_panel_style

        Select(self.selenium.find_element(By.ID, "theme_select")).select_by_value(
            "snow"
        )
        solar_panel_snow_style = self.selenium.find_element(
            By.ID, "solar_panel"
        ).get_attribute("style")
        assert "none" in solar_panel_snow_style

    def test_electric_fuel_gauge(self):
        self.login_once()

        page = self.go_to_level_editor()

        [road_start, road_end] = self.set_up_basic_map()
        page.go_to_character_tab()

        # the electric van has dropdown value 7 - select the electric van as character
        Select(self.selenium.find_element(By.ID, "character_select")).select_by_value(
            "7"
        )

        # save level and choose to play it
        page.go_to_save_tab()
        self.selenium.find_element(By.ID, "levelNameInput").send_keys(
            "test electric van level"
        )
        self.selenium.find_element(By.ID, "saveLevel").click()
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.visibility_of_element_located((By.ID, "myModal-lead"))
        )
        self.selenium.find_element(By.ID, "play_button").click()

        # check to see if electric fuel gauge appears
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.presence_of_element_located((By.ID, "electricFuelGauge"))
        )
        electric_fuel_gauge = self.selenium.find_element(By.ID, "electricFuelGauge")
        assert "visibility: visible" in electric_fuel_gauge.get_attribute("style")



# we should comment out languages that we don't have custom block translations for,
# although all these languages have built-in translations for the generic blocks

language_code_dict = {
  # "ab": "аҧсуа бызшәа",
  # "ar": "العربية",
  # "az": "azərbaycan",
  # "ba": "Башҡорт теле",
  # "bcc": "بلۏچی",
  # "be": "беларуская",
  # "be-tarask": "тарашкевіца",
  # "bg": "български",
  # "bn": "বাংলা",
  # "br": "Brezhoneg",
  "ca": "Català",
  # "cs": "čeština",
  # "da": "Dansk",
  # "de": "Deutsch",
  # "diq": "Zazakî",
  # "dty": "डोटेली",
  # "el": "Ελληνικά",
  "en": "English",
  "en-gb": "English (UK)",
  # "eo": "Esperanto",
  "es": "Español",
  # "et": "Eesti",
  # "eu": "Euskara",
  # "fa": "فارسی",
  # "fi": "Suomi",
  "fr": "Français",
  # "gl": "Galego",
  # "gor": "Hulontalo",
  # "ha": "Hausa",
  # "he": "עברית",
   "hi": "हिन्दी",
  # "hrx": "Hunsrik",
  # "hu": "Magyar",
  # "hy": "հայերեն",
  # "ia": "Interlingua",
  # "id": "Bahasa Indonesia",
  # "ig": "Igbo",
  # "is": "íslenska",
  # "it": "Italiano",
  # "ja": "日本語",
  # "kab": "Taqbaylit",
  # "ko": "한국어",
  # "lb": "Lëtzebuergesch",
  # "lki": "له‌کی",
  # "lo": "ລາວ",
  # "lrc": "لۊری شومالی",
  # "lt": "Lietuvių",
  # "lv": "Latviešu",
  # "mk": "македонски",
  # "mnw": "ဘာသာမန်",
  # "ms": "Bahasa Malaysia",
  # "nb": "Norsk bokmål",
  # "nl": "Nederlands",
  # "oc": "Occitan",
  # "pl": "Polski",
  # "pms": "Piemontèis",
  # "pt": "Português",
  # "pt-br": "Português (Brasil)",
  # "ro": "Română",
  # "ru": "русский",
  # "sc": "Sardu",
  # "sd": "سنڌي",
  # "shn": "လိၵ်ႈတႆး",
  # "sk": "Slovenčina",
  # "skr-arab": "سرائیکی",
  # "sl": "Slovenščina",
  # "sq": "Shqip",
  # "sr": "српски",
  # "sr-latn": "Srpski",
  # "sv": "Svenska",
  # "ta": "தமிழ்",
  # "tcy": "ತುಳು",
  # "te": "తెలుగు",
  # "th": "ไทย",
  # "tl": "Tagalog",
  # "tlh": "tlhIngan Hol",
  # "tr": "Türkçe",
  # "ug-arab": " ئۇيغۇرچە",
  # "uk": "українська",
  # "ur": "اردو",
  # "vi": "Tiếng Việt",
  # "xmf": "მარგალური ნინა",
  # "yo": "Èdè Yorùbá",
  # "zh-hans": "简体中文",
  # "zh-hant": "繁體中文"
}


import random
import string

from game.end_to_end_tests.base_game_test import BaseGameTest


class TestSavingWorkspace(BaseGameTest):
    already_logged_on = False
    user_profile = None

    def test_save_and_load_workspace(self):
        self.login_once()
        level = 10
        random_save_entry_name = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(10)
        )
        return (
            self.go_to_level(level)
            .save_solution(random_save_entry_name)
            .load_solution_by_name(random_save_entry_name)
        )



from django.apps.registry import Apps
from django.db import migrations


def disable_algo_score_for_levels_without_model_solution(apps: Apps, *args):
    Level = apps.get_model("game", "Level")

    Level.objects.filter(
        default=True, episode__isnull=False, model_solution="[]"
    ).update(disable_algorithm_score=True)

    Attempt = apps.get_model("game", "Attempt")

    Attempt.objects.filter(
        level__default=True,
        level__episode__isnull=False,
        level__model_solution="[]",
        score=20,
    ).update(score=10)


def enable_algo_score_for_levels_without_model_solution(apps: Apps, *args):
    Level = apps.get_model("game", "Level")

    Level.objects.filter(
        default=True, episode__isnull=False, model_solution="[]"
    ).update(disable_algorithm_score=False)

    Attempt = apps.get_model("game", "Attempt")

    Attempt.objects.filter(
        level__default=True,
        level__episode__isnull=False,
        level__model_solution="[]",
        score=10,
    ).update(score=20)


class Migration(migrations.Migration):
    dependencies = [("game", "0090_add_missing_model_solutions")]
    operations = [
        migrations.RunPython(
            disable_algo_score_for_levels_without_model_solution,
            reverse_code=enable_algo_score_for_levels_without_model_solution,
        )
    ]



from __future__ import absolute_import

from builtins import object

from rest_framework import serializers

from game import messages
from game.theme import get_theme, get_themes_url
from .models import Workspace, Level, Episode, LevelDecor, LevelBlock, Block


class WorkspaceSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = Workspace


class LevelListSerializer(serializers.HyperlinkedModelSerializer):
    title = serializers.SerializerMethodField()

    class Meta(object):
        model = Level
        fields = (
            "url",
            "__str__",
            "episode",
            "name",
            "title",
            "default",
            "blocklyEnabled",
            "pythonEnabled",
        )

    def get_title(self, obj):
        if obj.default:
            title = getattr(messages, "title_level" + obj.name)()
            return title
        else:
            return "Custom Level"


class LevelDetailSerializer(serializers.HyperlinkedModelSerializer):
    title = serializers.SerializerMethodField()
    description = serializers.SerializerMethodField()
    hint = serializers.SerializerMethodField()
    levelblock_set = serializers.HyperlinkedIdentityField(
        view_name="levelblock-for-level", read_only=True
    )
    map = serializers.HyperlinkedIdentityField(
        view_name="map-for-level", read_only=True
    )
    mode = serializers.HyperlinkedIdentityField(
        view_name="mode-for-level", read_only=True
    )

    class Meta(object):
        model = Level
        fields = (
            "__str__",
            "episode",
            "name",
            "title",
            "description",
            "hint",
            "next_level",
            "default",
            "levelblock_set",
            "map",
            "mode",
            "blocklyEnabled",
            "pythonEnabled",
            "pythonViewEnabled",
        )

    def get_title(self, obj):
        if obj.default:
            title = getattr(messages, "title_level" + obj.name)()
            return title
        else:
            return "Custom Level"

    def get_description(self, obj):
        return getattr(messages, "description_level" + obj.name)() if obj.default else obj.description

    def get_hint(self, obj):
        return getattr(messages, "hint_level" + obj.name)() if obj.default else obj.hint

    def get_leveldecor_set(self, obj):
        leveldecors = LevelDecor.objects.filter(level__id=obj.id)
        serializer = LevelDecorSerializer(
            leveldecors,
            many=True,
            context={"request": self.context.get("request", None)},
        )
        return serializer.data


class LevelModeSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = Level
        fields = ("blocklyEnabled", "pythonEnabled", "pythonViewEnabled")


class LevelMapListSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="map-for-level", read_only=True
    )

    class Meta(object):
        model = Level
        fields = ("url",)


class LevelMapDetailSerializer(serializers.HyperlinkedModelSerializer):
    leveldecor_set = serializers.SerializerMethodField()
    theme = serializers.SerializerMethodField()

    class Meta(object):
        model = Level
        fields = (
            "origin",
            "destinations",
            "path",
            "traffic_lights",
            "max_fuel",
            "theme",
            "leveldecor_set",
        )

    def get_leveldecor_set(self, obj):
        leveldecors = LevelDecor.objects.filter(level__id=obj.id)
        serializer = LevelDecorSerializer(
            leveldecors,
            many=True,
            context={"request": self.context.get("request", None)},
        )
        return serializer.data

    def get_theme(self, obj):
        pk = get_theme(obj.theme_name).pk
        return get_themes_url(pk, self.context.get("request", None))


class EpisodeListSerializer(serializers.HyperlinkedModelSerializer):
    name = serializers.SerializerMethodField()

    class Meta(object):
        model = Episode
        fields = ("url", "__str__", "name")

    def get_name(self, obj):
        return messages.get_episode_title(obj.id)


class EpisodeDetailSerializer(serializers.HyperlinkedModelSerializer):
    level_set = serializers.SerializerMethodField()
    level_set_url = serializers.HyperlinkedIdentityField(
        view_name="level-for-episode", read_only=True
    )
    name = serializers.SerializerMethodField()

    class Meta(object):
        model = Episode
        depth = 1
        fields = ("url", "__str__", "name", "level_set", "level_set_url")

    def get_level_set(self, obj):
        levels = Level.objects.filter(episode__id=obj.id)
        serializer = LevelListSerializer(
            levels, many=True, context={"request": self.context.get("request", None)}
        )
        return serializer.data

    def get_name(self, obj):
        return messages.get_episode_title(obj.id)


class LevelBlockSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = LevelBlock


class LevelDecorSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = LevelDecor


class BlockSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = Block
        fields = ("url", "id", "type")



from django.apps.registry import Apps
from django.db import migrations

def disable_algo_score_for_existing_custom_levels(apps: Apps, *args):
    Level = apps.get_model("game", "Level")

    Level.objects.filter(
        default=False
    ).update(disable_algorithm_score=True)

def enable_algo_score_for_existing_custom_levels(apps: Apps, *args):
    Level = apps.get_model("game", "Level")

    Level.objects.filter(
        default=False
    ).update(disable_algorithm_score=False)

class Migration(migrations.Migration):
    dependencies = [
        ("game", "0091_disable_algo_score_if_no_model_solution")
    ]

    operations = [
        migrations.RunPython(
            disable_algo_score_for_existing_custom_levels, 
            reverse_code=enable_algo_score_for_existing_custom_levels
        )
    ]


from __future__ import absolute_import

import itertools

from django import forms
from .models import Episode


class ScoreboardForm(forms.Form):
    def __init__(self, *args, **kwargs):
        classes = kwargs.pop("classes")
        super(ScoreboardForm, self).__init__(*args, **kwargs)
        classes_choices = [(c.id, c.name) for c in classes]

        self.fields["classes"] = forms.MultipleChoiceField(
            choices=classes_choices, widget=forms.CheckboxSelectMultiple()
        )
        # Each tuple in choices has two elements, id and name of each level
        # First element is the actual value set on the model
        # Second element is the string displayed on the dropdown menu
        episodes_choices = (
            (episode.id, episode.name)
            for episode in Episode.objects.filter(in_development=False)
        )
        self.fields["episodes"] = forms.MultipleChoiceField(
            choices=itertools.chain(episodes_choices),
            widget=forms.CheckboxSelectMultiple(),
        )


class LevelModerationForm(forms.Form):
    def __init__(self, *args, **kwargs):
        classes = kwargs.pop("classes")
        teacher = kwargs.pop("teacher")
        super(LevelModerationForm, self).__init__(*args, **kwargs)

        # If the teacher is an admin, append teacher names or "(you)" to classes
        if teacher.is_admin:
            classes_choices = [
                (
                    c.id,
                    f"{c.name} "
                    + (
                        "(you)"
                        if c.teacher == teacher
                        else f"({c.teacher.new_user.first_name} {c.teacher.new_user.last_name})"
                    ),
                )
                for c in classes
            ]
        else:
            classes_choices = [(c.id, c.name) for c in classes]

        self.fields["classes"] = forms.MultipleChoiceField(
            choices=classes_choices, widget=forms.CheckboxSelectMultiple()
        )



# Generated by Django 3.2.23 on 2023-12-19 21:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('game', '0086_loop_levels'),
    ]

    operations = [
        migrations.AddField(
            model_name='workspace',
            name='python_view_enabled',
            field=models.BooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("game", "0065_rename_old_character_field")]

    operations = [
        migrations.RemoveField(model_name="level", name="character_old"),
        migrations.DeleteModel(name="Character"),
    ]



from django.contrib import admin

from game.models import Level, Block, Episode, Workspace, LevelDecor, Attempt


class LevelAdmin(admin.ModelAdmin):
    search_fields = [
        "name",
        "id",
        "owner__user__username",
        "owner__user__first_name",
    ]
    raw_id_fields = ["next_level", "locked_for_class"]
    readonly_fields = ["owner"]
    list_display = ["name", "id", "episode", "owner"]


class EpisodeAdmin(admin.ModelAdmin):
    list_display = ["id", "name"]


class WorkspaceAdmin(admin.ModelAdmin):
    raw_id_fields = ["owner"]


class AttemptAdmin(admin.ModelAdmin):
    search_fields = [
        "level",
        "student",
        "start_time",
        "finish_time",
        "is_best_attempt",
    ]
    raw_id_fields = ["student"]
    list_display = [
        "level",
        "student",
        "start_time",
        "finish_time",
        "is_best_attempt",
    ]


class LevelDecorAdmin(admin.ModelAdmin):
    search_fields = ["level__name"]
    list_display = ["id", "level", "x", "y", "decorName"]


admin.site.register(Level, LevelAdmin)
admin.site.register(Episode, EpisodeAdmin)
admin.site.register(Workspace, WorkspaceAdmin)
admin.site.register(Block)
admin.site.register(Attempt, AttemptAdmin)
admin.site.register(LevelDecor, LevelDecorAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("game", "0061_auto_20160208_2144")]

    operations = [
        migrations.RemoveField(model_name="decor", name="theme"),
        migrations.RemoveField(model_name="level", name="theme_old"),
        migrations.DeleteModel(name="Decor"),
        migrations.DeleteModel(name="Theme"),
    ]



import json
from unittest.mock import patch

from common.models import Teacher
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from deploy import captcha
from django.core import mail
from django.test.client import Client
from django.test.testcases import TestCase
from django.urls import reverse
from hamcrest import assert_that, equal_to

from game.models import Level
from game.tests.utils.level import create_save_level, create_save_level_with_multiple_houses, multiple_house_data
from game.tests.utils.teacher import add_teacher_to_school, create_school


class LevelEditorTestCase(TestCase):
    LEVEL_DATA1 = {
        "origin": '{"coordinate":[3,5],"direction":"S"}',
        "pythonEnabled": False,
        "decor": [],
        "blocklyEnabled": True,
        "blocks": [{"type": "move_forwards"}, {"type": "turn_left"}, {"type": "turn_right"}],
        "max_fuel": "50",
        "pythonViewEnabled": False,
        "character": "3",
        "name": "abc",
        "theme": 1,
        "anonymous": False,
        "cows": "[]",
        "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],' '"connectedNodes":[0]}]',
        "traffic_lights": "[]",
        "destinations": "[[3,4]]",
    }

    @classmethod
    def setUpClass(cls):
        cls.orig_captcha_enabled = captcha.CAPTCHA_ENABLED
        captcha.CAPTCHA_ENABLED = False
        super(LevelEditorTestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        captcha.CAPTCHA_ENABLED = cls.orig_captcha_enabled
        super(LevelEditorTestCase, cls).tearDownClass()

    def setUp(self):
        self.client = Client()

    def login(self, email, password):
        self.client.post(
            reverse("teacher_login"),
            {"auth-username": email, "auth-password": password, "teacher_login_view-current_step": "auth"},
            follow=True,
        )

    def logout(self):
        self.client.post(reverse("logout_view"), follow=True)

    def student_login(self, name, access_code, password):
        self.client.post(
            reverse("student_login", kwargs={"access_code": access_code}),
            {"username": name, "password": password},
            follow=True,
        )

    def get_sharing_information(self, level_id):
        url = reverse("get_sharing_information_for_editor", args=[level_id])
        response = self.client.get(url)
        return response

    def share_level_for_editor(self, level_id):
        url = reverse("share_level_for_editor", args=[level_id])
        response = self.client.post(url)
        return response

    def test_level_saving_school_student(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, class_name, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.student_login(student_name, access_code, student_password)
        url = reverse("save_level_for_editor")
        response = self.client.post(url, {"data": json.dumps(self.LEVEL_DATA1)})

        assert_that(response.status_code, equal_to(200))
        sharing_info1 = json.loads(self.get_sharing_information(json.loads(response.content)["id"]).getvalue())
        assert sharing_info1["teacher"]["shared"]
        assert len(mail.outbox) == 1

    def test_anonymous_level_saving_school_student(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, class_name, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.student_login(student_name, access_code, student_password)
        url = reverse("save_level_for_editor")
        data1 = {
            "origin": '{"coordinate":[3,5],"direction":"S"}',
            "pythonEnabled": False,
            "decor": [],
            "blocklyEnabled": True,
            "blocks": [{"type": "move_forwards"}, {"type": "turn_left"}, {"type": "turn_right"}],
            "max_fuel": "50",
            "pythonViewEnabled": False,
            "character": "3",
            "name": "abc",
            "theme": 1,
            "anonymous": True,
            "cows": "[]",
            "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],' '"connectedNodes":[0]}]',
            "traffic_lights": "[]",
            "destinations": "[[3,4]]",
        }
        response = self.client.post(url, {"data": json.dumps(data1)})

        assert response.status_code == 200
        sharing_info1 = json.loads(self.get_sharing_information(json.loads(response.content)["id"]).getvalue())
        assert sharing_info1["teacher"]["shared"]
        assert len(mail.outbox) == 0

    def test_level_sharing_with_no_school(self):
        email1, password1 = signup_teacher_directly()
        teacher1 = Teacher.objects.get(new_user__email=email1)
        self.login(email1, password1)
        level = create_save_level(teacher1)

        sharing_info1 = json.loads(self.get_sharing_information(level.id).getvalue())
        assert len(sharing_info1["teachers"]) == 0

    def test_level_sharing_with_school(self):
        email1, password1 = signup_teacher_directly()
        email2, _ = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)
        teacher2 = Teacher.objects.get(new_user__email=email2)

        self.login(email1, password1)
        level = create_save_level(teacher1)

        school1 = create_school()
        add_teacher_to_school(teacher1, school1)
        add_teacher_to_school(teacher2, school1)

        sharing_info1 = json.loads(self.get_sharing_information(level.id).getvalue())
        assert len(sharing_info1["teachers"]) == 1

    def test_level_sharing_with_empty_school(self):
        email1, password1 = signup_teacher_directly()
        teacher1 = Teacher.objects.get(new_user__email=email1)

        self.login(email1, password1)
        level = create_save_level(teacher1)

        school1 = create_school()
        add_teacher_to_school(teacher1, school1)

        sharing_info1 = json.loads(self.get_sharing_information(level.id).getvalue())
        assert len(sharing_info1["teachers"]) == 0

    def test_level_sharing_permissions(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)
        teacher2 = Teacher.objects.get(new_user__email=email2)

        self.login(email1, password1)
        level = create_save_level(teacher1)
        share_url = reverse("share_level_for_editor", args=[level.id])

        school1 = create_school()
        add_teacher_to_school(teacher1, school1, is_admin=True)
        add_teacher_to_school(teacher2, school1)

        # Create a class and 2 students for the second teacher
        _, class_name2, access_code2 = create_class_directly(email2)
        student_name1, student_password1, student1 = create_school_student_directly(access_code2)
        student_name2, student_password2, student2 = create_school_student_directly(access_code2)

        self.logout()
        self.login(email2, password2)

        # Second teacher can't share the level as it's not been shared with them yet
        response = self.client.post(share_url, {"recipientIDs[]": [student1.new_user.id], "action": ["share"]})
        assert response.status_code == 403

        # Log in as the first teacher again
        self.logout()
        self.login(email1, password1)

        # Share the level with the second teacher
        response = self.client.post(share_url, {"recipientIDs[]": [teacher2.new_user.id], "action": ["share"]})
        assert response.status_code == 200

        # Log in as the second teacher
        self.logout()
        self.login(email2, password2)

        # Now the second teacher should be able to share the level
        response = self.client.post(share_url, {"recipientIDs[]": [student1.new_user.id], "action": ["share"]})
        assert response.status_code == 200
        # and load it
        load_level_url = reverse("load_level_for_editor", kwargs={"levelID": level.id})
        response = self.client.get(load_level_url)
        assert response.status_code == 200

        # Login as the first student
        self.logout()
        self.student_login(student_name1, access_code2, student_password1)

        # Check that the student cannot share the level
        sharing_info = json.loads(self.get_sharing_information(level.id).getvalue())
        assert sharing_info["detail"] == "You do not have permission to perform this action."

        # Check the student can view the level
        response = self.client.get(reverse("play_custom_level", args=[level.id]))
        assert response.status_code == 200

        # Login as first teacher again
        self.logout()
        self.login(email1, password1)

        # Share the level with the second student of teacher 2
        response = self.client.post(share_url, {"recipientIDs[]": [student2.new_user.id], "action": ["share"]})
        assert response.status_code == 200

        # Login as the second student
        self.logout()
        self.student_login(student_name2, access_code2, student_password2)

        # Check that the student cannot share the level
        sharing_info = json.loads(self.get_sharing_information(level.id).getvalue())
        assert sharing_info["detail"] == "You do not have permission to perform this action."

        # Check the student can view the level
        response = self.client.get(reverse("play_custom_level", args=[level.id]))
        assert response.status_code == 200

        self.logout()

    def test_level_can_only_be_edited_by_owner(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, student = create_school_student_directly(access_code)
        hacker_name, hacker_password, hacker = create_school_student_directly(access_code)

        new_level = Level.objects.create(owner_id=student.user_id, name="my_custom_level")

        self.student_login(hacker_name, access_code, hacker_password)
        resp = self.client.get(reverse("level_editor_chosen_level", kwargs={"levelId": new_level.id}))

        assert "level" not in resp.context

        self.logout()

        self.student_login(student_name, access_code, student_password)
        resp = self.client.get(reverse("level_editor_chosen_level", kwargs={"levelId": new_level.id}))

        assert "level" in resp.context

    def test_no_character_set_defaults_to_van(self):

        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, class_name, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.student_login(student_name, access_code, student_password)
        url = reverse("save_level_for_editor")
        data_with_no_character = {
            "origin": '{"coordinate":[3,5],"direction":"S"}',
            "pythonEnabled": False,
            "decor": [],
            "blocklyEnabled": True,
            "blocks": [{"type": "move_forwards"}, {"type": "turn_left"}, {"type": "turn_right"}],
            "max_fuel": "50",
            "pythonViewEnabled": False,
            "name": "abc",
            "theme": 1,
            "anonymous": True,
            "cows": "[]",
            "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],' '"connectedNodes":[0]}]',
            "traffic_lights": "[]",
            "destinations": "[[3,4]]",
        }
        response = self.client.post(url, {"data": json.dumps(data_with_no_character)})

        assert response.status_code == 200
        new_level = Level.objects.get(name="abc")
        assert new_level.character.name == "Van"

    def test_language_set_appropriately(self):

        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, class_name, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.student_login(student_name, access_code, student_password)
        url = reverse("save_level_for_editor")
        data_with_split_language = {
            "origin": '{"coordinate":[3,5],"direction":"S"}',
            "pythonEnabled": False,
            "decor": [],
            "blocklyEnabled": True,
            "blocks": [
                {"type": "move_forwards"},
                {"type": "turn_left"},
                {"type": "turn_right"},
            ],
            "max_fuel": "50",
            "pythonViewEnabled": True,
            "name": "abc",
            "theme": 1,
            "anonymous": True,
            "cows": "[]",
            "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],'
            '"connectedNodes":[0]}]',
            "traffic_lights": "[]",
            "destinations": "[[3,4]]",
        }
        response = self.client.post(url, {"data": json.dumps(data_with_split_language)})

        assert response.status_code == 200
        new_level = Level.objects.get(name="abc")

        assert new_level.pythonViewEnabled
        assert not new_level.pythonEnabled
        assert new_level.blocklyEnabled

    def test_level_loading(self):
        email1, password1 = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)

        self.login(email1, password1)
        level = create_save_level(teacher1)
        url = reverse("load_level_for_editor", kwargs={"levelID": level.id})
        response = self.client.get(url)

        assert response.status_code == 200

    def test_level_saving_with_multiple_houses(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, class_name, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.student_login(student_name, access_code, student_password)
        url = reverse("save_level_for_editor")
        data_with_multiple_houses = {
            "origin": '{"coordinate":[3,5],"direction":"S"}',
            "pythonEnabled": False,
            "decor": [],
            "blocklyEnabled": True,
            "blocks": [{"type": "move_forwards"}, {"type": "turn_left"}, {"type": "turn_right"}],
            "max_fuel": "50",
            "pythonViewEnabled": False,
            "name": "multiple_houses",
            "theme": 1,
            "anonymous": True,
            "cows": "[]",
            "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],"connectedNodes":[0,2]}, {"coordinate":[3,3],"connectedNodes":[1]}]',
            "traffic_lights": "[]",
            "destinations": "[[3,4],[3,3]]",
        }
        response = self.client.post(url, {"data": json.dumps(data_with_multiple_houses)})

        assert response.status_code == 200
        new_level = Level.objects.get(name="multiple_houses")
        assert new_level.destinations == "[[3,4],[3,3]]"

    def test_level_loading_with_multiple_houses(self):

        email1, password1 = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)

        self.login(email1, password1)
        level = create_save_level_with_multiple_houses(teacher1)
        url = reverse("load_level_for_editor", kwargs={"levelID": level.id})
        response = self.client.get(url)
        response_data = response.json()

        assert response.status_code == 200
        assert response_data["level"]["destinations"] == "[[3,4], [3,3]]"

    @patch("game.level_management.save_level")
    def test_custom_level_scoring(self, mock_save_level):
        email1, password1 = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)

        self.login(email1, password1)

        level = create_save_level_with_multiple_houses(teacher1)

        save_url = reverse("save_level_for_editor", kwargs={"levelId": level.id})
        response = self.client.post(save_url, {"data": json.dumps(multiple_house_data)})

        disable_algorithm_score = mock_save_level.call_args.args[1]["disable_algorithm_score"]

        assert response.status_code == 200
        assert disable_algorithm_score

    def test_level_of_anonymised_teacher_is_hidden(self):
        # Create 2 teacher accounts
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)
        teacher2 = Teacher.objects.get(new_user__email=email2)

        # Add teachers to the same school
        school = create_school()
        add_teacher_to_school(teacher1, school)
        add_teacher_to_school(teacher2, school)

        # Create a level as `teacher2` and share it with `teacher1`
        self.login(email2, password2)
        create_save_level(teacher2, shared_with=[teacher1.new_user])

        # Log in as `teacher1`
        self.logout()
        self.login(email1, password1)

        # Check `teacher1` can see `teacher2`'s shared level
        levels_url = reverse("levels")
        response = self.client.get(levels_url)
        assert len(response.context["directly_shared_levels"]) == 1

        # Make teacher2 inactive
        teacher2.new_user.is_active = 0
        teacher2.new_user.save()

        # `teacher1` shouldn't see any shared levels now
        response = self.client.get(levels_url)
        assert len(response.context["directly_shared_levels"]) == 0

    def test_level_of_anonymised_student_is_hidden(self):
        # Create a teacher and a student
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, student = create_school_student_directly(access_code)

        teacher = Teacher.objects.get(new_user__email=email)

        # Create a level as student and share it with teacher
        self.student_login(student_name, access_code, student_password)
        create_save_level(student, shared_with=[teacher.new_user])

        # Log in as teacher
        self.logout()
        self.login(email, password)

        # Check teacher can see student's shared level
        levels_url = reverse("levels")
        response = self.client.get(levels_url)
        assert len(response.context["directly_shared_levels"]) == 1

        # Make student inactive
        student.new_user.is_active = 0
        student.new_user.save()

        # Teacher shouldn't see any shared levels now
        response = self.client.get(levels_url)
        assert len(response.context["directly_shared_levels"]) == 0



# Generated by Django 3.2.25 on 2024-06-11 14:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('game', '0092_disable_algo_score_in_custom_levels'),
    ]

    operations = [
        migrations.AlterField(
            model_name='level',
            name='character_name',
            field=models.CharField(blank=True, choices=[('Van', 'Van'), ('Dee', 'Dee'), ('Nigel', 'Nigel'), ('Kirsty', 'Kirsty'), ('Wes', 'Wes'), ('Phil', 'Phil'), ('Electric van', 'Electric van')], default=None, max_length=20, null=True),
        ),
    ]



from django.apps.registry import Apps
from django.db import migrations


def mark_episodes_in_development(apps: Apps, *args):
    Episode = apps.get_model("game", "Episode")

    for i in range(13, 16):
        episode = Episode.objects.get(pk=i)
        episode.in_development = True
        episode.save()


def unmark_episodes_in_development(apps: Apps, *args):
    Episode = apps.get_model("game", "Episode")

    for i in range(13, 16):
        episode = Episode.objects.get(pk=i)
        episode.in_development = False
        episode.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0088_rename_episodes")]
    operations = [
        migrations.RunPython(
            mark_episodes_in_development,
            reverse_code=unmark_episodes_in_development,
        )
    ]



import datetime
import logging
import re

import pytz
from django.apps.registry import Apps
from django.db import migrations

LOGGER = logging.getLogger(__name__)


def add_missing_model_solutions(apps: Apps, *args):
    Level = apps.get_model("game", "Level")

    model_solutions = {
        "80": "[5]",
        "81": "[10]",
        "82": "[11]",
        "83": "[5]",
        "84": "[5]",
        "85": "[3]",
        "86": "[8]",
        "87": "[8]",
        "88": "[11]",
        "89": "[11]",
        "90": "[15]",
        "91": "[15]",
    }

    for level_name, model_solution in model_solutions.items():
        count = Level.objects.filter(
            default=True, episode__isnull=False, name=level_name
        ).update(model_solution=model_solution, disable_algorithm_score=False)
        assert count == 1

    Attempt = apps.get_model("game", "Attempt")

    attempts = Attempt.objects.filter(
        level__default=True,
        level__episode__isnull=False,
        level__name__in=[
            "80",
            "81",
            "82",
            "83",
            "84",
            "85",
            "86",
            "87",
            "88",
            "89",
            "90",
            "91",
        ],
        score=10,
        finish_time__gte=datetime.datetime(2023, 3, 1, 00, 00, tzinfo=pytz.UTC),
    ).select_related("level")

    LOGGER.info(f"Retrieved {attempts.count()} attempts.")

    processed_count = 0

    for attempt in attempts.iterator(chunk_size=1000):
        workspace = attempt.workspace

        # Get number of blocks from solution - 1 to discount Start block
        number_of_blocks = len(re.findall("<block", workspace)) - 1

        model_solution = model_solutions[attempt.level.name]

        ideal_number_of_blocks = int(
            model_solution.replace("[", "").replace("]", "")
        )

        if number_of_blocks == ideal_number_of_blocks:
            attempt.score = 20
        else:
            difference = abs(number_of_blocks - ideal_number_of_blocks)

            if difference < 10:
                attempt.score += 10 - difference

        attempt.save()

        processed_count += 1

        if processed_count % 1000 == 0:
            LOGGER.info(f"Processed {processed_count} attempts.")


def remove_new_model_solutions(apps: Apps, *args):
    Level = apps.get_model("game", "Level")

    Level.objects.filter(
        default=True,
        episode__isnull=False,
        name__in=[
            "80",
            "81",
            "82",
            "83",
            "84",
            "85",
            "86",
            "87",
            "88",
            "89",
            "90",
            "91",
        ],
    ).update(model_solution="[]", disable_algorithm_score=True)

    Attempt = apps.get_model("game", "Attempt")

    Attempt.objects.filter(
        level__default=True,
        level__episode__isnull=False,
        level__name__in=[
            "80",
            "81",
            "82",
            "83",
            "84",
            "85",
            "86",
            "87",
            "88",
            "89",
            "90",
            "91",
        ],
        score__gt=10,
        finish_time__gte=datetime.datetime(2023, 3, 1, 00, 00, tzinfo=pytz.UTC),
    ).select_related("level").update(score=10)


class Migration(migrations.Migration):
    dependencies = [("game", "0089_episodes_in_development")]
    operations = [
        migrations.RunPython(
            add_missing_model_solutions,
            reverse_code=remove_new_model_solutions,
        )
    ]



from django.apps.registry import Apps
from django.db import migrations


def rename_episodes(apps: Apps, *args):
    Episode = apps.get_model("game", "Episode")

    def update_episode_name(pk: int, name: str):
        episode = Episode.objects.get(pk=pk)
        episode.name = name
        episode.save()

    update_episode_name(pk=12, name="Counted Loops Using While")
    update_episode_name(pk=13, name="Indeterminate While Loops - coming soon")


def undo_rename_episodes(apps: Apps, *args):
    Episode = apps.get_model("game", "Episode")

    def update_episode_name(pk: int, name: str):
        episode = Episode.objects.get(pk=pk)
        episode.name = name
        episode.save()

    update_episode_name(pk=12, name="Sequencing and Counted Loops")
    update_episode_name(pk=13, name="Indeterminate WHILE Loops - coming soon")

class Migration(migrations.Migration):
    dependencies = [("game", "0087_workspace_python_view_enabled")]
    operations = [
        migrations.RunPython(
            rename_episodes,
            reverse_code=undo_rename_episodes,
        )
    ]



from __future__ import division

from django.shortcuts import render


def renderError(request, title, message):
    """Renders an error page with passed title and message.

    **Context**

    ``RequestContext``
    ``title``
        Title that is to be used as a title and header of the page. String.
    ``message``
        Message that will be shown on the error page. String.

    **Template:**

    :template:`game/error.html`
    """
    return render(
        request, "game/error.html", context={"title": title, "message": message}
    )



from __future__ import division

import json
import re
from builtins import map
from builtins import str

from common.models import Student, Teacher
from django.contrib.auth.models import User
from django.db import transaction
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, render, redirect
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_POST
from portal.templatetags import app_tags
from rest_framework.authentication import SessionAuthentication
from rest_framework.views import APIView

import game.level_management as level_management
import game.messages as messages
import game.permissions as permissions
from game import app_settings
from game import random_road
from game.cache import cached_level_decor, cached_level_blocks
from game.character import get_all_character
from game.decor import get_all_decor, get_decor_element
from game.models import Level, Block, UserProfile
from game.theme import get_all_themes
from game.views.level import LevelSerializer


def level_editor(request, levelId=None):
    """Renders the level editor page.

    **Context**

    ``RequestContext``
    ``blocks``
        Blocks that can be chosen to be played with later on.
        List of :model:`game.Block`.

    **Template:**

    :template:`game/level_editor.html`
    """

    context = {
        "blocks": [block.type for block in available_blocks()],
        "decor": get_all_decor(),
        "characters": get_all_character(),
        "themes": get_all_themes(),
        "cow_level_enabled": app_settings.COW_FEATURE_ENABLED,
        "night_mode_feature_enabled": str(
            app_settings.NIGHT_MODE_FEATURE_ENABLED
        ).lower(),
    }
    if levelId:
        level = Level.objects.get(id=levelId)
        user_profile = UserProfile.objects.get(id=level.owner_id)

        if request.user.id == user_profile.user_id:
            context["level"] = levelId

    return render(request, "game/level_editor.html", context=context)


def available_blocks():
    if app_settings.COW_FEATURE_ENABLED:
        return Block.objects.all()
    else:
        return Block.objects.all().exclude(type__in=["cow_crossing", "sound_horn"])


def play_anonymous_level(request, levelId, from_level_editor=True, random_level=False):
    night_mode = (
        False if not app_settings.NIGHT_MODE_FEATURE_ENABLED else "night" in request.GET
    )
    level = Level.objects.filter(id=levelId)

    if not level.exists():
        return redirect(reverse("level_editor"), permanent=True)

    level = level[:1].get()

    if not level.anonymous:
        return redirect(reverse("level_editor"), permanent=True)

    subtitle = level.subtitle
    lesson = level.lesson
    hint = level.hint

    attempt = None
    house = get_decor_element("house", level.theme).url
    cfc = get_decor_element("cfc", level.theme).url
    background = get_decor_element("tile1", level.theme).url
    character = level.character

    character_url = character.top_down
    character_width = character.width
    character_height = character.height
    wreckage_url = "van_wreckage.svg"

    decor_data = cached_level_decor(level)

    if night_mode:
        block_data = level_management.get_night_blocks(level)
        night_mode = "true"
        lesson = messages.title_night_mode()
        model_solution = "[]"
    else:
        block_data = cached_level_blocks(level)
        night_mode = "false"
        model_solution = level.model_solution

    return_view_name = "level_editor" if from_level_editor else "levels"

    level.delete()

    return render(
        request,
        "game/game.html",
        context={
            "level": level,
            "decor": decor_data,
            "blocks": block_data,
            "subtitle": subtitle,
            "lesson": lesson,
            "hint": hint,
            "character": character,
            "background": background,
            "house": house,
            "cfc": cfc,
            "attempt": attempt,
            "random_level": random_level,
            "return_url": reverse(return_view_name),
            "character_url": character_url,
            "character_width": character_width,
            "character_height": character_height,
            "wreckage_url": wreckage_url,
            "night_mode": night_mode,
            "night_mode_feature_enabled": str(
                app_settings.NIGHT_MODE_FEATURE_ENABLED
            ).lower(),
            "model_solution": model_solution,
        },
    )


def level_data_for(level):
    return {
        "name": level.name,
        "owner": app_tags.make_into_username(level.owner.user),
        "id": level.id,
    }


def levels_shared_with(user):
    shared_levels = level_management.levels_shared_with(user)
    shared_data = list(map(level_data_for, shared_levels))
    return shared_data


def levels_owned_by(user):
    levels_owned_by_user = level_management.levels_owned_by(user)
    owned_data = list(map(level_data_for, levels_owned_by_user))
    return owned_data


def get_list_of_loadable_levels(user):
    owned_data = levels_owned_by(user)
    shared_data = levels_shared_with(user)

    return {"ownedLevels": owned_data, "sharedLevels": shared_data}


def owned_levels(request):
    level_data = levels_owned_by(request.user)
    return HttpResponse(json.dumps(level_data), content_type="application/javascript")


def shared_levels(request):
    level_data = levels_shared_with(request.user)
    return HttpResponse(json.dumps(level_data), content_type="application/javascript")


def get_loadable_levels_for_editor(request):
    response = get_list_of_loadable_levels(request.user)
    return HttpResponse(json.dumps(response), content_type="application/javascript")


def load_level_for_editor(request, levelID):
    level = get_object_or_404(Level, id=levelID)

    if not permissions.can_load_level(request.user, level):
        return HttpResponseUnauthorized()

    level_dict = LevelSerializer(level).data
    level_dict["theme"] = level.theme.id
    level_dict["decor"] = level_management.get_decor(level)
    level_dict["blocks"] = cached_level_blocks(level)

    response = {"owned": level.owner == request.user.userprofile, "level": level_dict}

    return HttpResponse(json.dumps(response), content_type="application/json")


@transaction.atomic
@require_POST
def save_level_for_editor(request, levelId=None):
    """Processes a request on creation of the map in the level editor"""
    data = json.loads(request.POST["data"])
    data["disable_algorithm_score"] = True

    if ("character" not in data) or (not data["character"]):
        # Set a default, to deal with issue #1158 "Cannot save custom level"
        data["character"] = 1
    if levelId is not None:
        level = get_object_or_404(Level, id=levelId)
    else:
        level = Level(default=False, anonymous=data["anonymous"])
        if permissions.can_create_level(request.user):
            level.owner = request.user.userprofile
    if not permissions.can_save_level(request.user, level):
        return HttpResponseUnauthorized()

    pattern = re.compile("^(\w?[ ]?)*$")

    if pattern.match(data["name"]):
        level_management.save_level(level, data)

        if levelId is None:
            teacher = None

            is_user_school_student = (
                hasattr(level.owner, "student")
                and not level.owner.student.is_independent()
            )
            is_user_independent = (
                hasattr(level.owner, "student") and level.owner.student.is_independent()
            )
            is_user_teacher = hasattr(level.owner, "teacher")

            # if level owner is a school student, share with teacher automatically if they aren't an admin
            if is_user_school_student:
                teacher = level.owner.student.class_field.teacher
                if not teacher.is_admin:
                    level.shared_with.add(teacher.new_user)

                if not data["anonymous"]:
                    level_management.email_new_custom_level(
                        level.owner.student.class_field.teacher.new_user.email,
                        request.build_absolute_uri(reverse("level_moderation")),
                        request.build_absolute_uri(
                            reverse("play_custom_level", kwargs={"levelId": level.id})
                        ),
                        request.build_absolute_uri(reverse("home")),
                        str(level.owner.student),
                        level.owner.student.class_field.name,
                    )
            elif is_user_teacher:
                teacher = level.owner.teacher

            # share with all admins of the school if user is in a school
            if not is_user_independent:
                school_admins = teacher.school.admins()

                [
                    level.shared_with.add(school_admin.new_user)
                    for school_admin in school_admins
                    if school_admin.new_user != request.user
                ]

            level.save()
        response = {"id": level.id}
        return HttpResponse(json.dumps(response), content_type="application/json")
    else:
        return HttpResponseUnauthorized()


@transaction.atomic
def delete_level_for_editor(request, levelId):
    level = get_object_or_404(Level, id=levelId)

    if not permissions.can_delete_level(request.user, level):
        return HttpResponseUnauthorized()

    level_management.delete_level(level)

    return HttpResponse(json.dumps({}), content_type="application/javascript")


@require_POST
def generate_random_map_for_editor(request):
    """
    Generates a new random path suitable for a random level with the parameters provided
    """
    data = dict(request.POST)

    size = int(data["numberOfTiles"][0])
    branchiness = float(data["branchiness"][0])
    loopiness = float(data["loopiness"][0])
    curviness = float(data["curviness"][0])
    traffic_lights = data["trafficLights"][0] == "true"
    scenery = data["scenery"][0] == "true"

    data = random_road.generate_random_map_data(
        size, branchiness, loopiness, curviness, traffic_lights, scenery, False
    )

    return HttpResponse(json.dumps(data), content_type="application/javascript")


class SharingInformationForEditor(APIView):
    """Returns information about who the level can be and is shared with. This uses
    the CanShareLevel permission."""

    authentication_classes = (SessionAuthentication,)
    permission_classes = (permissions.CanShareLevel,)

    def get(self, request, **kwargs):
        """
        Gets a level's information of who it is and can be shared with. How it works:
        - if the requester is a student, get all their classmates and their teacher.
        - if the requester is a teacher, get all their students and their colleagues.
        :param request: GET request made by user.
        :param kwargs: In this case, houses the level ID.
        :return: A HttpResponse with the data in JSON format.
        """
        levelID = kwargs["levelID"]
        level = get_object_or_404(Level, id=levelID)

        self.check_object_permissions(request, level)

        userprofile = request.user.userprofile
        valid_recipients = {}

        # Note: independent users can't share levels so no need to check
        if hasattr(userprofile, "student"):
            student = userprofile.student

            # First get all the student's classmates
            class_ = student.class_field
            classmates = Student.objects.filter(
                class_field=class_, new_user__is_active=True
            ).exclude(id=student.id)
            valid_recipients["classmates"] = [
                {
                    "id": classmate.new_user.id,
                    "name": app_tags.make_into_username(classmate.new_user),
                    "shared": level.owner == classmate.user
                    or level.shared_with.filter(id=classmate.new_user.id).exists(),
                }
                for classmate in classmates
            ]

            # Then add their teacher as well
            teacher = class_.teacher
            valid_recipients["teacher"] = {
                "id": teacher.new_user.id,
                "name": app_tags.make_into_username(teacher.new_user),
                "shared": level.owner == teacher.user
                or level.shared_with.filter(id=teacher.new_user.id).exists(),
            }

        elif hasattr(userprofile, "teacher"):
            teacher = userprofile.teacher

            # First get all the students they teach
            valid_recipients["classes"] = []
            if teacher.is_admin:
                classes_taught = teacher.school.classes()
            else:
                classes_taught = teacher.class_teacher.all()
            for class_ in classes_taught:
                students = Student.objects.filter(
                    class_field=class_, new_user__is_active=True
                )
                valid_recipients["classes"].append(
                    {
                        "name": f"{class_.name} ({app_tags.make_into_username(class_.teacher.new_user)})"
                        if teacher.is_admin
                        else class_.name,
                        "id": class_.id,
                        "students": [
                            {
                                "id": student.new_user.id,
                                "name": app_tags.make_into_username(student.new_user),
                                "shared": level.owner == student.user
                                or level.shared_with.filter(
                                    id=student.new_user.id
                                ).exists(),
                            }
                            for student in students
                        ],
                    }
                )

            if not teacher.school:
                valid_recipients["teachers"] = []
            else:
                fellow_teachers = Teacher.objects.filter(school=teacher.school)
                valid_recipients["teachers"] = [
                    {
                        "id": fellow_teacher.new_user.id,
                        "name": app_tags.make_into_username(fellow_teacher.new_user),
                        "admin": fellow_teacher.is_admin,
                        "shared": level.owner == fellow_teacher.user
                        or level.shared_with.filter(
                            id=fellow_teacher.new_user.id
                        ).exists(),
                    }
                    for fellow_teacher in fellow_teachers
                    if teacher != fellow_teacher
                ]

        return HttpResponse(
            json.dumps(valid_recipients), content_type="application/javascript"
        )


class ShareLevelView(APIView):
    """Handles the sharing request of a level."""

    authentication_classes = (SessionAuthentication,)
    permission_classes = [permissions.CanShareLevel, permissions.CanShareLevelWith]

    def post(self, request, **kwargs):
        """
        Gets the level id and recipient ids from the request, and shares or
        unshares the level according to the action from the request.
        :param request: the post request sent to the view.
        :return a call to the SharingInformationForEditor class to return the new
        sharing information of the level.
        """
        levelID = kwargs["levelID"]
        recipientIDs = request.POST.getlist("recipientIDs[]")
        action = request.POST.get("action")

        level = get_object_or_404(Level, id=levelID)

        recipients = User.objects.filter(id__in=recipientIDs)
        users = self._get_users_to_share_level_with(recipients)

        if action == "share":
            level_management.share_level(level, *users)
        elif action == "unshare":
            level_management.unshare_level(level, *users)

        return SharingInformationForEditor().get(request, levelID=levelID)

    def _get_users_to_share_level_with(self, recipients):
        """
        Gets a list of users that the level can be shared with - this is done by
        checking the list of requested users against the sharing permission.
        :param recipients: List of recipients the user wants to share the level with.
        :return: A list of users which the requester is authorised to share the level
        with.
        """
        return [
            recipient.userprofile.user
            for recipient in recipients
            if self._can_share_level_with(recipient)
        ]

    def _can_share_level_with(self, recipient):
        """
        Checks whether the requester can share a level with a specific user. Calls the
        CanShareLevelWith permission.
        :param recipient: User that the requester wants to share a level with.
        :return: A boolean of whether the requester has permission.
        """
        return permissions.CanShareLevelWith().has_object_permission(
            self.request, self, recipient
        )


class HttpResponseUnauthorized(HttpResponse):
    def __init__(self):
        super(HttpResponseUnauthorized, self).__init__(
            content="Unauthorized", status=401
        )



from __future__ import unicode_literals

from builtins import map, range
from datetime import datetime, timedelta

from common.models import Class, Teacher, Student
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import (
    create_school_student_directly,
    create_independent_student_directly,
)
from common.tests.utils.teacher import signup_teacher_directly
from django.test import Client, TestCase
from django.urls import reverse

from game.models import Attempt, Level, Episode
from game.tests.utils.level import create_save_level
from game.views.scoreboard import (
    StudentRow,
    scoreboard_data,
    Headers,
    StudentInTrouble,
    shared_level_to_name,
    shared_levels_data,
    SharedHeaders,
)
from game.views.scoreboard_csv import (
    scoreboard_csv,
    Headers as CSVHeaders,
    SharedHeaders as CSVSharedHeaders,
)


class ScoreboardTestCase(TestCase):
    def test_teacher_multiple_students_multiple_levels(self):
        # Setup official levels data
        episode_ids = [1, 2]
        episode1 = Episode.objects.get(id=1)
        episode2 = Episode.objects.get(id=2)
        level_ids = [
            f"{x}"
            for x in range(1, len(episode1.levels) + len(episode2.levels) + 1)
        ]
        level1 = Level.objects.get(name="1")
        level13 = Level.objects.get(name="13")

        clas, student, student2 = set_up_data()

        create_attempt(student, level1, 20)
        create_attempt(student2, level1, 2)
        create_attempt(student2, level13, 16)

        # Setup custom levels data
        shared_level = create_save_level(
            student, "custom_level1", shared_with=[student2.new_user]
        )

        create_attempt(student2, shared_level, 10)

        all_levels = [level1, level13]
        all_shared_levels = [shared_level]

        attempts_per_student = {
            student: Attempt.objects.filter(
                level__in=all_levels, student=student, is_best_attempt=True
            ).select_related("level"),
            student2: Attempt.objects.filter(
                level__in=all_levels, student=student2, is_best_attempt=True
            ).select_related("level"),
        }

        shared_attempts_per_student = {
            student2: Attempt.objects.filter(
                level__in=all_shared_levels,
                student=student2,
                is_best_attempt=True,
            ).select_related("level"),
        }

        # Generate results
        student_data, headers, level_headers, levels_sorted = scoreboard_data(
            episode_ids, attempts_per_student
        )
        (
            shared_headers,
            shared_level_headers,
            shared_student_data,
        ) = shared_levels_data(
            student.new_user, all_shared_levels, shared_attempts_per_student
        )

        # Check data for official levels matches
        assert headers == Headers
        assert level_headers == [f"L{id}" for id in level_ids]

        assert len(student_data) == 2

        student_row = student_data[0]
        assert student_row.class_field.name == clas.name
        assert student_row.name == student.user.user.first_name
        assert student_row.total_score == 20
        assert student_row.total_time == timedelta(0)
        assert student_row.level_scores[all_levels[0].id]["score"] == 20
        assert student_row.completed == 1
        assert student_row.success_rate == 100.0

        student_row = student_data[1]
        assert student_row.class_field.name == clas.name
        assert student_row.name == student2.user.user.first_name
        assert student_row.total_score == 18
        assert student_row.total_time == timedelta(0)
        assert student_row.level_scores[all_levels[0].id]["score"] == 2
        assert student_row.level_scores[all_levels[1].id]["score"] == 16
        assert student_row.completed == 1
        assert (
            student_row.success_rate == 45.0
        )  ## the scores, (2 + 16 = 18), divided by the total possible, (2 * 20 = 40), 18/40 = 45%

        # Check data for custom levels matches
        assert shared_headers == SharedHeaders
        assert shared_level_headers == [
            f"{shared_level.name} ({shared_level.owner})"
        ]

        assert len(shared_student_data) == 1

        student_row = shared_student_data[0]
        assert student_row.class_field.name == clas.name
        assert student_row.name == student2.user.user.first_name
        assert student_row.total_score == 10
        assert student_row.total_time == timedelta(0)
        assert student_row.level_scores[shared_level.id]["score"] == 10
        assert student_row.completed == 1
        assert student_row.success_rate == 100.0

    def test_scoreboard_loads(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, name, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        url = reverse("scoreboard")
        c = Client()
        c.login(username=email, password=password)

        # test scoreboard page loads properly
        response = c.get(url)
        assert response.status_code == 200

        # test scoreboard shows all episodes if no episodes are manually selected
        data = {"classes": [klass.id], "view": [""]}

        response = c.post(url, data)

        active_levels = Level.objects.filter(episode__in_development=False)

        assert response.status_code == 200
        assert len(response.context["level_headers"]) == active_levels.count()

    def test_student_can_see_classes(self):
        """A student should be able to see the classes they are in"""
        mr_teacher = Teacher.objects.factory(
            "Normal", "Teacher", "normal@school.edu", "secretpa$sword"
        )
        klass, name1, _ = create_class_directly(
            mr_teacher.user.user.email, class_name="Class 1"
        )
        _, name2, _ = create_class_directly(
            mr_teacher.user.user.email, class_name="Class 2"
        )
        student = Student.objects.schoolFactory(klass, "some student", "secret")

        c = Client()
        c.force_login(student.user.user)

        url = reverse("scoreboard")
        response = c.get(url)

        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]
        assert name1 in choices_in_form
        assert name2 not in choices_in_form
        assert len(choices_in_form) == 1

    def test_admin_teacher_can_see_all_classes(self):
        """An admin should be able to see all classes, not just the ones they teach"""
        normal_teacher = Teacher.objects.factory(
            "Normal", "Teacher", "normal@school.edu", "secretpa$sword"
        )
        admin_teacher = Teacher.objects.factory(
            "Admin", "Admin", "admin@school.edu", "secretpa$sword2"
        )

        admin_teacher.is_admin = True
        admin_teacher.save()

        _, name1, _ = create_class_directly(
            admin_teacher.user.user.email, class_name="Class 1"
        )
        _, name2, _ = create_class_directly(
            admin_teacher.user.user.email, class_name="Class 2"
        )
        _, name3, _ = create_class_directly(
            normal_teacher.user.user.email, class_name="Class 3"
        )

        c = Client()
        c.login(
            username=admin_teacher.user.user.email, password="secretpa$sword2"
        )

        url = reverse("scoreboard")
        response = c.get(url)

        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]

        assert name1 in choices_in_form
        assert name2 in choices_in_form
        assert name3 in choices_in_form

    def test_non_admin_teacher_can_only_see_their_own_classes(self):
        """A teacher who is not an admin should only be able to see their classes, not ones taught by others"""
        teacher1 = Teacher.objects.factory(
            "First", "Teacher", "normal@school.edu", "secretpa$sword"
        )
        teacher2 = Teacher.objects.factory(
            "Second", "Teacher", "admin@school.edu", "secretpa$sword2"
        )

        _, name1, _ = create_class_directly(
            teacher2.user.user.email, class_name="Class 1"
        )
        _, name2, _ = create_class_directly(
            teacher2.user.user.email, class_name="Class 2"
        )
        _, name3, _ = create_class_directly(
            teacher1.user.user.email, class_name="Class 3"
        )

        c = Client()
        # First teacher logs in.  Should see only Class 3
        c.login(username="normal@school.edu", password="secretpa$sword")

        url = reverse("scoreboard")
        response = c.get(url)

        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]

        assert name3 in choices_in_form
        assert name1 not in choices_in_form
        assert name2 not in choices_in_form
        assert len(choices_in_form) == 1

        # Other teacher logs in.  Should see Class 1 and Class 2
        c.login(username="admin@school.edu", password="secretpa$sword2")

        response = c.get(url)
        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]

        assert name3 not in choices_in_form
        assert name1 in choices_in_form
        assert name2 in choices_in_form
        assert len(choices_in_form) == 2

    def test_independent_student_cannot_see_scoreboard(self):
        username, password, _ = create_independent_student_directly()

        c = Client()
        c.login(username=username, password=password)

        url = reverse("scoreboard")
        response = c.get(url)

        assert (
            "Scoreboard is only visible to school students and teachers"
            in str(response.content)
        )


class ScoreboardCsvTestCase(TestCase):
    def test_scoreboard_csv(self):
        # Take the first two levels of the main game
        levels = Level.objects.sorted_levels()[0:2]

        # Create 2 students along with their main scoreboard data
        student_rows = [None, None]
        students = [None, None]
        student_rows[0], students[0] = self.student_row("secrète")
        student_rows[1], students[1] = self.student_row()

        # Create 2 custom levels and create the associated student data
        shared_level_rows = [None, None]
        shared_level1 = create_save_level(
            students[0], "level1", shared_with=[students[1].new_user]
        )
        shared_level2 = create_save_level(students[1], "level2")
        shared_levels = [shared_level1, shared_level2]

        shared_levels_headers = list(
            [
                shared_level_to_name(level, level.owner)
                for level in shared_levels
            ]
        )

        shared_level_rows[0] = self.shared_student_row(
            students[0], shared_levels
        )
        shared_level_rows[1] = self.shared_student_row(
            students[1], shared_levels
        )

        # Create students' improvement table data
        improvement_data = []
        for student in students:
            stud = StudentInTrouble(student=student, areas="Getting started")
            improvement_data.append(stud)

        # Generate the CSV
        response = scoreboard_csv(
            student_rows,
            levels,
            improvement_data,
            shared_levels_headers,
            shared_level_rows,
        )

        # Gather the data from the CSV
        (
            actual_scoreboard_header,
            actual_scoreboard_rows,
            actual_shared_levels_header,
            actual_shared_levels_rows,
        ) = self.actual_data(response.content.decode("utf-8"), len(students))

        # Check the headers and the number or rows match expectations
        assert actual_scoreboard_header == self.expected_scoreboard_header(
            levels
        )
        assert len(actual_scoreboard_rows) == len(student_rows)
        assert (
            actual_shared_levels_header
            == self.expected_shared_levels_header(shared_levels)
        )
        assert len(actual_shared_levels_rows) == len(shared_level_rows)

        # check first scoreboard row
        (
            class_name,
            name,
            completed_levels,
            total_time,
            total_scores,
            l1,
            l2,
            improvement,
        ) = actual_scoreboard_rows[0].split(",")
        assert student_rows[0].class_field.name == class_name
        assert student_rows[0].name == name
        assert student_rows[0].level_scores[0]["score"] == int(l1)
        assert student_rows[0].level_scores[1]["score"] == int(l2)
        assert improvement == "Getting started"

        # check last scoreboard row
        last = len(actual_scoreboard_rows) - 1
        (
            class_name,
            name,
            completed_levels,
            total_time,
            total_scores,
            l1,
            l2,
            improvement,
        ) = actual_scoreboard_rows[last].split(",")
        assert student_rows[last].class_field.name == class_name
        assert student_rows[last].name == name
        assert str(student_rows[last].total_time) == total_time

        # check first shared level row
        (class_name, name, l1, l2) = actual_shared_levels_rows[0].split(",")
        assert shared_level_rows[0].class_field.name == class_name
        assert shared_level_rows[0].name == name
        assert shared_level_rows[0].level_scores[0]["score"] == l1
        assert shared_level_rows[0].level_scores[1]["score"] == l2

        # check last shared level row
        last = len(actual_shared_levels_rows) - 1
        (class_name, name, l1, l2) = actual_shared_levels_rows[last].split(",")
        assert shared_level_rows[last].class_field.name == class_name
        assert shared_level_rows[last].name == name
        assert shared_level_rows[last].level_scores[0]["score"] == int(l1)
        assert shared_level_rows[last].level_scores[1]["score"] == l2

    def student_row(self, class_name=None):
        """
        Create data for a student row in the main scoreboard table
        """
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email, class_name)
        _, _, student = create_school_student_directly(access_code)

        total_time = timedelta(seconds=30)
        scores = [x for x in range(20)]
        total_score = sum(scores)

        level_scores = {}
        for i in range(2):
            level_scores[i] = {}
            level_scores[i]["score"] = 5
            level_scores[i]["full_score"] = 20
            level_scores[i]["is_low_attempt"] = True

        row = StudentRow(
            student=student,
            class_field=Class(name="MyClass"),
            total_time=total_time,
            total_score=total_score,
            level_scores=level_scores,
            completed=2,
            success_rate=45,
        )

        return row, student

    def shared_student_row(self, student, shared_levels):
        """
        Create data for a student row in the custom levels table
        """
        level_scores = {}
        for i in range(len(shared_levels)):
            level_scores[i] = {}

            level_scores[i]["score"] = 5

            if student.new_user not in shared_levels[i].shared_with.all():
                level_scores[i]["score"] = "Not shared"

            if shared_levels[i].owner == student.user:
                level_scores[i]["score"] = "Owner"

            level_scores[i]["full_score"] = 10
            level_scores[i]["is_low_attempt"] = True

        row = StudentRow(
            student=student,
            class_field=Class(name="MyClass"),
            total_time=0,
            total_score=0,
            level_scores=level_scores,
            completed=0,
            success_rate=0,
        )

        return row

    def expected_scoreboard_header(self, levels):
        level_strings = list(map(str, levels))
        all_header_strings = (
            CSVHeaders + level_strings + ["Areas for improvement"]
        )
        joined = ",".join(all_header_strings)
        return joined

    def expected_shared_levels_header(self, shared_levels):
        level_strings = list(
            [
                shared_level_to_name(level, level.owner)
                for level in shared_levels
            ]
        )
        all_header_strings = CSVSharedHeaders + level_strings
        joined = ",".join(all_header_strings)
        return joined

    def actual_data(self, content, number_of_students):
        """
        Gets the data from the CSV taking into account the number of students in the request. The CSV should contain
        two tables, each with a one-row header and x number of rows where x = number of students.
        """
        scoreboard_header_row = 0
        scoreboard_rows_start = scoreboard_header_row + 1
        scoreboard_rows_end = scoreboard_rows_start + number_of_students
        shared_levels_header_row = scoreboard_rows_end + 1
        shared_levels_rows_start = shared_levels_header_row + 1
        shared_levels_rows_end = shared_levels_rows_start + number_of_students

        split = content.strip().split("\r\n")
        scoreboard_header = split[scoreboard_header_row]
        scoreboard_rows = split[scoreboard_rows_start:scoreboard_rows_end]
        shared_levels_header = split[shared_levels_header_row]
        shared_levels_rows = split[
            shared_levels_rows_start:shared_levels_rows_end
        ]

        return (
            scoreboard_header,
            scoreboard_rows,
            shared_levels_header,
            shared_levels_rows,
        )


def create_attempt(student, level, score):
    attempt = Attempt.objects.create(
        finish_time=datetime.fromtimestamp(1435305072),
        level=level,
        student=student,
        score=score,
        is_best_attempt=True,
    )
    attempt.start_time = datetime.fromtimestamp(1435305072)
    attempt.save()


def ids_of_levels_named(names):
    levels = Level.objects.filter(name__in=names)
    return levels


def set_up_data(classmates_data_viewable=False):
    email, password = signup_teacher_directly()
    clas, class_name, access_code = create_class_directly(email)
    if classmates_data_viewable:
        clas.classmates_data_viewable = True
        clas.save()
    _, _, student = create_school_student_directly(access_code)
    _, _, student2 = create_school_student_directly(access_code)
    create_random_school_data()
    return clas, student, student2


def create_random_school_data():
    email, password = signup_teacher_directly()
    clas, class_name, access_code = create_class_directly(email)
    create_school_student_directly(access_code)
    create_school_student_directly(access_code)



from game.end_to_end_tests.base_game_test import BaseGameTest


class TestTurningAround(BaseGameTest):
    def test_turn_around_on_straight_road(self):
        self.running_out_of_instructions_test(
            level=40, workspace_file="turn_around_on_straight_road"
        )

    def test_turn_around_on_left_turn(self):
        self.running_out_of_instructions_test(
            level=40, workspace_file="turn_around_on_left_turn"
        )

    def test_turn_around_on_right_turn(self):
        self.running_out_of_instructions_test(
            level=40, workspace_file="turn_around_on_right_turn"
        )
